/*
 * Copyright(c) 2013 Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Contact Information:
 * Intel Corporation
 */

#include <netboot/linux-asm-io.h>
#include <stage2/shared.h>
#include "intel_quark_sb.h"

#define INTEL_QUARK_SB_CMD_ADDR	(0x000000D0)
#define INTEL_QUARK_SB_DATA_ADDR	(0x000000D4)

#define INTEL_QUARK_SB_MCR_SHIFT	(24)
#define INTEL_QUARK_SB_PORT_SHIFT	(16)
#define INTEL_QUARK_SB_REG_SHIFT	(8)
#define INTEL_QUARK_SB_BYTEEN	(0xF0)	/* enable all 32 bits */

/* PCI config space reg definitions */
#define PCI_CLASS_DEVICE	(0x0A)

struct sb_pci_dev {
  unsigned int bus;
  unsigned int dev_fn;
};

static struct sb_pci_dev sb_pcidev;

/*
 * Functions for accessing PCI configuration space with type 1 accesses
 */
#define CONFIG_CMD(bus, dev_fn, where) \
  (0x80000000 | (bus << 16) | (dev_fn << 8) | (where & ~3))

static void
pci_read_config_byte(struct sb_pci_dev *sb_pcidev,
                     unsigned int where, unsigned char *value)
{
  outl(CONFIG_CMD(sb_pcidev->bus,sb_pcidev->dev_fn, where), 0xCF8);
  *value = inb(0xCFC + (where&3));
}

static void
pci_read_config_word (struct sb_pci_dev *sb_pcidev,
                      unsigned int where, unsigned short *value)
{
  outl(CONFIG_CMD(sb_pcidev->bus,sb_pcidev->dev_fn,where), 0xCF8);
  *value = inw(0xCFC + (where&2));
}

static void
pci_read_config_dword (struct sb_pci_dev *sb_pcidev,
                       unsigned int where, unsigned int *value)
{
  outl(CONFIG_CMD(sb_pcidev->bus,sb_pcidev->dev_fn, where), 0xCF8);
  *value = inl(0xCFC);
}

static void
pci_write_config_byte (struct sb_pci_dev *sb_pcidev,
                       unsigned int where, unsigned char value)
{
  outl(CONFIG_CMD(sb_pcidev->bus,sb_pcidev->dev_fn, where), 0xCF8);
  outb(value, 0xCFC + (where&3));
}

static void
pci_write_config_word (struct sb_pci_dev *sb_pcidev,
                       unsigned int where, unsigned short value)
{
  outl(CONFIG_CMD(sb_pcidev->bus,sb_pcidev->dev_fn, where), 0xCF8);
  outw(value, 0xCFC + (where&2));
}

static void
pci_write_config_dword (struct sb_pci_dev *sb_pcidev, unsigned int where,
                        unsigned int value)
{
  outl(CONFIG_CMD(sb_pcidev->bus,sb_pcidev->dev_fn, where), 0xCF8);
  outl(value, 0xCFC);
}

/**
 * intel_quark_sb_read_reg
 *
 * @param quark_sb_id: Sideband identifier
 * @param command: Command to send to destination identifier
 * @param reg: Target register w/r to quark_sb_id
 * @return nothing
 *
 * Utility function to allow thread-safe read of side-band
 * command - can be different read op-code types - which is why we don't
 * hard-code this value directly into msg
 */
void
intel_quark_sb_read_reg(quark_sb_id id, grub_uint8_t cmd, grub_uint8_t reg,
                      grub_uint32_t *data)
{
  grub_uint32_t msg = (cmd << INTEL_QUARK_SB_MCR_SHIFT) |
      ((id << INTEL_QUARK_SB_PORT_SHIFT) & 0xFF0000)|
      ((reg << INTEL_QUARK_SB_REG_SHIFT) & 0xFF00)|
      INTEL_QUARK_SB_BYTEEN;

  if(data == NULL)
    return;

  pci_write_config_dword(&sb_pcidev, INTEL_QUARK_SB_CMD_ADDR, msg);
  pci_read_config_dword(&sb_pcidev, INTEL_QUARK_SB_DATA_ADDR, data);

}

/**
 * intel_quark_sb_write_reg
 *
 * @param quark_sb_id: Sideband identifier
 * @param command: Command to send to destination identifier
 * @param reg: Target register w/r to quark_sb_id
 * @return nothing
 *
 * Utility function to allow thread-safe write of side-band
 */
void
intel_quark_sb_write_reg(quark_sb_id id, grub_uint8_t cmd, grub_uint8_t reg,
                       grub_uint32_t data)
{
  grub_uint32_t msg = (cmd << INTEL_QUARK_SB_MCR_SHIFT) |
      ((id << INTEL_QUARK_SB_PORT_SHIFT) & 0xFF0000)|
      ((reg << INTEL_QUARK_SB_REG_SHIFT) & 0xFF00)|
      INTEL_QUARK_SB_BYTEEN;

  pci_write_config_dword(&sb_pcidev, INTEL_QUARK_SB_DATA_ADDR, data);
  pci_write_config_dword(&sb_pcidev, INTEL_QUARK_SB_CMD_ADDR, msg);
}

/**
 * intel_quark_sb_probe
 *
 * @param found_devid: the found PCI device ID (if any)
 * @return 0 if success, !0 otherwise
 */
int
intel_quark_sb_probe(grub_uint16_t *found_devid)
{
  int found = 0;
  grub_uint16_t class;
  grub_uint16_t device, vendor;

  sb_pcidev.bus = 0;
  for (sb_pcidev.dev_fn = 0; sb_pcidev.dev_fn < 0xFF; sb_pcidev.dev_fn++)
    {
      /* Only probe function 0 on single fn devices */
      pci_read_config_word(&sb_pcidev, PCI_CLASS_DEVICE, &class);

      if (class == 0xffff)
        continue;

      pci_read_config_word(&sb_pcidev, PCI_VENDOR_ID, &vendor);
      pci_read_config_word(&sb_pcidev, PCI_DEVICE_ID, &device);

      if(vendor != PCI_VENDOR_ID_INTEL)
        continue;

      if (device == PCI_DEVICE_ID_QUARK_X1000_HBRIDGE || device == 0x12C0)
        {
          found = 1;
          *found_devid = device;
          if (grub_quark_debug)
            grub_printf("Host Bridge B/DF 0x%04x/0x%04x VID/DID 0x%04x/0x%04x\n",
                        sb_pcidev.bus, sb_pcidev.dev_fn, vendor, device);
          return 0;
        }
    }

  grub_printf("Couldn't enumerate Host Bridge\n");
  errnum = ERR_DEV_VALUES;

  return errnum;
}
